/*
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 *                          http://www.ntop.org
 *
 * Copyright (C) 2004 Burton Strauss <Burton@ntopSupport.com>
 *
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* #define DEBUG */

#define DNSCACHEDB                          "dnsCache.db"

#include <gdbm.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "globals-defines.h"

// Extracted from #include "globals-structtypes.h"
typedef struct storedAddress {
  char   symAddress[MAX_LEN_SYM_HOST_NAME];
  time_t recordCreationTime;
  short  symAddressType;
} StoredAddress;

#undef CONST_DNSCACHE_LIFETIME
#define CONST_DNSCACHE_LIFETIME 5400

int main(int argc, char *argv[]) {
  int aCount, dCount, kCount, nCount, oCount, pCount, rCount, sCount, dpCount, rc;

  GDBM_FILE dnsCacheFile;
  datum key, nextkey, db_data;

  struct stat statbuf;

  char timeBuf[LEN_TIMEFORMAT_BUFFER];
  char fileBuf[128];

  struct tm t;
  time_t now = time(NULL);
  time_t purge = time(NULL) - CONST_DNSCACHE_LIFETIME;

  struct in6_addr junk;
  unsigned int u_junk;

  aCount=0;
  dCount=0;
  kCount=0;
  nCount=0;
  oCount=0;
  pCount=0;
  sCount=0;
  dpCount=1; /* Fake 'successful' prior pass... */

  printf("dnsCachePurge - Burton Strauss - Burton@ntopSupport.com\n\n");

  if((argc != 2) || 
     (argv[1][0] == '-')) {
      printf("\tUsage:\n\t\tdnsCachePurge directory\n\n"
             "\tThis program opens the ntop " DNSCACHEDB " gdbm database in exclusive\n"
             "\twrite mode (i.e. ntop must NOT be running) and purges bad records.\n\n"
             "\tThe filename is <directory>/" DNSCACHEDB ".\n\n"
             "\tBad records are defined as:\n\n"
             "\t\t1. Key is not a valid address - neither IPv4 (as an unsigned number) or\n"
             "\t\t   IPv6 (any of the inet_pton formats).\n"
             "\t\t2. Record is the wrong size (missing it's time stamp or type flag).\n"
             "\t\t3. Record is older than the CONST_DNSCACHE_LIFETIME value specified in\n"
             "\t\t   globals-defines.h.\n\n"
             "\tBecause of the record read algorithm (see man gdbm), it is not possible\n"
             "\tto count the input records.  Further, since this could miss a records, we\n"
             "\tmake up to five passes through the file, stopping after a pass in which we\n"
             "\tdo not delete any records\n"
             "\n\n");
      exit(0);
  }

  strftime(timeBuf, sizeof(timeBuf), "%c", localtime_r(&now, &t));
  snprintf(fileBuf, sizeof(fileBuf), "%s/%s", argv[1], DNSCACHEDB);
  printf("dnsCachePurge of %s at %s\n", fileBuf, timeBuf);

  if(stat(fileBuf, &statbuf) != 0) {
      printf("...unable to find database!\n");
      exit(1);
  }

  dnsCacheFile = gdbm_open (fileBuf, 0, GDBM_WRITER, 00664, NULL);

  if(dnsCacheFile == NULL) {
      printf("...open failed: %s\n", gdbm_strerror(gdbm_errno));
      exit(1);
  }

  strftime(timeBuf, sizeof(timeBuf), "%c", localtime_r(&purge, &t));
  printf("\trecords older than %s\n", timeBuf);

  printf("\nRunning...\n");

  while (dpCount != 0) {
      ++pCount;
      rCount=0;

      if(pCount > 5) {
          printf("\n\nToo many passes... stopping!\n\n");
          break;
      }

      printf("\tpass %d...", pCount);
#ifdef DEBUG
      printf("\n");
#endif

      dpCount = 0;
      oCount = 0;

      memset(&key, 0, sizeof(key));

      key = gdbm_firstkey(dnsCacheFile);
      while(key.dptr != NULL) {
          ++rCount;
          nextkey = gdbm_nextkey(dnsCacheFile, key);

          rc = inet_pton(AF_INET6, key.dptr, &junk);
          if(rc <= 0) {
              rc = sscanf(key.dptr, "%u", &u_junk);
              /* 1 from sscanf means ok (1 item assigned).  0 or EOF is a problem... */
          }

          if(rc <= 0) {
#ifdef DEBUG
              printf("%-20s\n", key.dptr);
#endif
              ++kCount;
              ++dCount;
              ++dpCount;
              gdbm_delete(dnsCacheFile, key);
              free(key.dptr);
              key = nextkey;
              continue;
          }

          db_data = gdbm_fetch(dnsCacheFile, key);

          if(db_data.dptr == NULL) {
                  ++nCount;
                  ++dCount;
                  ++dpCount;
                  gdbm_delete(dnsCacheFile, key);
           } else {

#ifdef DEBUG
              strftime(timeBuf, sizeof(timeBuf), "%m/%d %H:%M",
                       localtime_r(&((StoredAddress*)db_data.dptr)->recordCreationTime, &t));
              printf("%-20s %11s %3d %s\n",
                     key.dptr,
                     timeBuf,
                     ((StoredAddress*)db_data.dptr)->symAddressType,
                     ((StoredAddress*)db_data.dptr)->symAddress);
#endif

              if(db_data.dsize != (sizeof(StoredAddress)+1)) {
                  ++sCount;
                  ++dCount;
                  ++dpCount;
                  gdbm_delete(dnsCacheFile, key);
              } else if(((StoredAddress*)db_data.dptr)->recordCreationTime < purge) {
                  ++aCount;
                  ++dCount;
                  ++dpCount;
                  gdbm_delete(dnsCacheFile, key);
              } else {
                  ++oCount;
              }

              free(db_data.dptr);

          }

          free(key.dptr);

          key = nextkey;
      }
      printf("%d records read %d deleted\n", rCount, dpCount);
  }

  gdbm_close(dnsCacheFile);

  printf("\n\nDone\n\n"
         "%5d bad key\n"
         "%5d No data\n"
         "%5d wrong size\n"
         "%5d too old\n"
         "-----"\n"
         "%5d Gives: records deleted\n"
         "%5d records out (approx.)\n\n",
         dCount,
         kCount,
         nCount,
         sCount,
         aCount,
         oCount);
}
       
